from django.urls import path
from . import views

urlpatterns = [
  path('test/', views.test),
  path('empty/', views.empty_test),
  path('redirect/', views.redirect_test),
  path('props/', views.props_test),
  path('template_data/', views.template_data_test),
  path('lazy/', views.lazy_test),
  path('optional/', views.optional_test),
  path('defer/', views.defer_test),
  path('defer-group/', views.defer_group_test),
  path('merge/', views.merge_test),
  path('complex-props/', views.complex_props_test),
  path('share/', views.share_test),
  path('inertia-redirect/', views.inertia_redirect_test),
  path('external-redirect/', views.external_redirect_test),
  path('encrypt-history/', views.encrypt_history_test),
  path('no-encrypt-history/', views.encrypt_history_false_test),
  path('encrypt-history-type-error/', views.encrypt_history_type_error_test),
  path('clear-history/', views.clear_history_test),
  path('clear-history-redirect/', views.clear_history_redirect_test),
  path('clear-history-type-error/', views.clear_history_type_error_test),
]



from inertia.test import InertiaTestCase

class TestTestCase(InertiaTestCase):

  def test_include_props(self):
    response = self.client.get('/props/')

    self.assertIncludesProps({'name': 'Brandon'})

  def test_has_exact_props(self):
    response = self.client.get('/props/')

    self.assertHasExactProps({'name': 'Brandon', 'sport': 'Hockey'})

  def test_has_template_data(self):
    response = self.client.get('/template_data/')

    self.assertIncludesTemplateData({'name': 'Brian'})

  def test_has_exact_template_data(self):
    response = self.client.get('/template_data/')

    self.assertHasExactTemplateData({'name': 'Brian', 'sport': 'Basketball'})

  def test_component_name(self):
    response = self.client.get('/props/')

    self.assertComponentUsed('TestComponent')




from inertia.tests.testapp.models import User
from django.test import TestCase
from inertia.utils import InertiaJsonEncoder
from json import dumps
from datetime import date, datetime

class InertiaJsonEncoderTestCase(TestCase):
  def setUp(self):
    self.encode = lambda obj: dumps(obj, cls=InertiaJsonEncoder)

  def test_it_handles_models_with_dates_and_removes_passwords(self):
    user = User(
      name='Brandon',
      password='something-top-secret',
      birthdate=date(1987, 2, 15),
      registered_at=datetime(2022, 10, 31, 10, 13, 1),
    )

    self.assertEqual(
      dumps({'id': None, 'name': 'Brandon', 'birthdate': '1987-02-15', 'registered_at': '2022-10-31T10:13:01'}),
      self.encode(user)
    )

  def test_it_handles_querysets(self):
    User(
      name='Brandon',
      password='something-top-secret',
      birthdate=date(1987, 2, 15),
      registered_at=datetime(2022, 10, 31, 10, 13, 1),
    ).save()

    self.assertEqual(
      dumps([{'id': 1, 'name': 'Brandon', 'birthdate': '1987-02-15', 'registered_at': '2022-10-31T10:13:01'}]),
      self.encode(User.objects.all())
    )


from django.http.response import HttpResponse
from django.shortcuts import redirect
from django.utils.decorators import decorator_from_middleware
from inertia import inertia, render, lazy, merge, optional, defer, share, location
from inertia.http import INERTIA_SESSION_CLEAR_HISTORY, clear_history, encrypt_history

class ShareMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response

  def process_request(self, request):
    share(request, 
      position=lambda: 'goalie',
      number=29,
    )

def test(request):
  return HttpResponse('Hey good stuff')

@inertia('TestComponent')
def empty_test(request):
  return {}

def redirect_test(request):
  return redirect(empty_test)

@inertia('TestComponent')
def inertia_redirect_test(request):
  return redirect(empty_test)

def external_redirect_test(request):
  return location("http://foobar.com/")

@inertia('TestComponent')
def props_test(request):
  return {
    'name': 'Brandon',
    'sport': 'Hockey',
  }

def template_data_test(request):
  return render(request, 'TestComponent', template_data={
    'name': 'Brian',
    'sport': 'Basketball',
  })

@inertia('TestComponent')
def lazy_test(request):
  return {
    'name': 'Brian',
    'sport': lazy(lambda: 'Basketball'),
    'grit': lazy(lambda: 'intense'),
  }

@inertia('TestComponent')
def optional_test(request):
  return {
    'name': 'Brian',
    'sport': optional(lambda: 'Basketball'),
    'grit': optional(lambda: 'intense'),
  }

@inertia('TestComponent')
def defer_test(request):
  return {
    'name': 'Brian',
    'sport': defer(lambda: 'Basketball')
  }


@inertia('TestComponent')
def defer_group_test(request):
  return {
    'name': 'Brian',
    'sport': defer(lambda: 'Basketball', 'group'),
    'team': defer(lambda: 'Bulls', 'group'),
    'grit': defer(lambda: 'intense')
  }

@inertia('TestComponent')
def merge_test(request):
  return {
    'name': 'Brandon',
    'sport': merge(lambda: 'Hockey'),
    'team': defer(lambda: 'Penguins', merge=True),
  }

@inertia('TestComponent')
def complex_props_test(request):
  return {
    'person': {
      'name': lambda: 'Brandon',
    }
  }

@decorator_from_middleware(ShareMiddleware)
@inertia('TestComponent')
def share_test(request):
  return {
    'name': 'Brandon',
  }

@inertia('TestComponent')
def encrypt_history_test(request):
  encrypt_history(request)
  return {}

@inertia('TestComponent')
def encrypt_history_false_test(request):
  encrypt_history(request, False)
  return {}

@inertia('TestComponent')
def encrypt_history_type_error_test(request):
  encrypt_history(request, "foo")
  return {}

@inertia('TestComponent')
def clear_history_test(request):
  clear_history(request)
  return {}

@inertia('TestComponent')
def clear_history_redirect_test(request):
  clear_history(request)
  return redirect(empty_test)

@inertia('TestComponent')
def clear_history_type_error_test(request):
  request.session[INERTIA_SESSION_CLEAR_HISTORY] = "foo"
  return {}



from http import HTTPStatus
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render as base_render
from .settings import settings
from json import dumps as json_encode
from functools import wraps
import requests
from .prop_classes import IgnoreOnFirstLoadProp, DeferredProp, MergeableProp

INERTIA_REQUEST_ENCRYPT_HISTORY = "_inertia_encrypt_history"
INERTIA_SESSION_CLEAR_HISTORY = "_inertia_clear_history"

def render(request, component, props={}, template_data={}):
  def is_a_partial_render():
    return 'X-Inertia-Partial-Data' in request.headers and request.headers.get('X-Inertia-Partial-Component', '') == component

  def partial_keys():
    return request.headers.get('X-Inertia-Partial-Data', '').split(',')

  def deep_transform_callables(prop):
    if not isinstance(prop, dict):
      return prop() if callable(prop) else prop
    
    for key in list(prop.keys()):
      prop[key] = deep_transform_callables(prop[key])

    return prop

  def build_props():
    _props = {
      **(request.inertia.all() if hasattr(request, 'inertia') else {}),
      **props,
    }

    for key in list(_props.keys()):
      if is_a_partial_render():
        if key not in partial_keys():
          del _props[key]
      else:
        if isinstance(_props[key], IgnoreOnFirstLoadProp):
          del _props[key]

    return deep_transform_callables(_props)

  def build_deferred_props():
    if is_a_partial_render():
      return None

    _deferred_props = {}
    for key, prop in props.items():
      if isinstance(prop, DeferredProp):
        _deferred_props.setdefault(prop.group, []).append(key)

    return _deferred_props
  
  def build_merge_props():
    reset_keys = request.headers.get('X-Inertia-Reset', '').split(',')

    return [
      key 
      for key, prop in props.items()
      if (
        isinstance(prop, MergeableProp)
        and prop.should_merge()
        and key not in reset_keys
      )
    ]
    
  def render_ssr():
    data = json_encode(page_data(), cls=settings.INERTIA_JSON_ENCODER)
    response = requests.post(
      f"{settings.INERTIA_SSR_URL}/render",
      data=data,
      headers={"Content-Type": "application/json"},
    )
    response.raise_for_status()
    return base_render(request, 'inertia_ssr.html', {
     'inertia_layout': settings.INERTIA_LAYOUT,
      **response.json()
    })

  def page_data():
    encrypt_history = getattr(request, INERTIA_REQUEST_ENCRYPT_HISTORY, settings.INERTIA_ENCRYPT_HISTORY)
    if not isinstance(encrypt_history, bool):
      raise TypeError(f"Expected boolean for encrypt_history, got {type(encrypt_history).__name__}")

    clear_history = request.session.pop(INERTIA_SESSION_CLEAR_HISTORY, False)
    if not isinstance(clear_history, bool):
      raise TypeError(f"Expected boolean for clear_history, got {type(clear_history).__name__}")

    _page = {
      'component': component,
      'props': build_props(),
      'url': request.build_absolute_uri(),
      'version': settings.INERTIA_VERSION,
      'encryptHistory': encrypt_history,
      'clearHistory': clear_history,
    }

    _deferred_props = build_deferred_props()
    if _deferred_props:
      _page['deferredProps'] = _deferred_props
    
    _merge_props = build_merge_props()
    if _merge_props:
      _page['mergeProps'] = _merge_props
    
    return _page

  if 'X-Inertia' in request.headers:
    return JsonResponse(
      data=page_data(),
      headers={
        'Vary': 'X-Inertia',
        'X-Inertia': 'true',
      },
      encoder=settings.INERTIA_JSON_ENCODER,
    )

  if settings.INERTIA_SSR_ENABLED:
    try:
      return render_ssr()
    except Exception:
      pass
  
  return base_render(request, 'inertia.html', {
    'inertia_layout': settings.INERTIA_LAYOUT,
    'page': json_encode(page_data(), cls=settings.INERTIA_JSON_ENCODER),
    **template_data,
  })

def location(location):
    return HttpResponse('', status=HTTPStatus.CONFLICT, headers={
      'X-Inertia-Location': location,
    })

def encrypt_history(request, value=True):
  setattr(request, INERTIA_REQUEST_ENCRYPT_HISTORY, value)

def clear_history(request):
  request.session[INERTIA_SESSION_CLEAR_HISTORY] = True

def inertia(component):
  def decorator(func):
    @wraps(func)
    def inner(request, *args, **kwargs):
      props = func(request, *args, **kwargs)

      # if something other than a dict is returned, the user probably wants to return a specific response
      if not isinstance(props, dict):
        return props

      return render(request, component, props)
    
    return inner
  
  return decorator



from django.db import models

class User(models.Model):
  name = models.CharField(max_length=255)
  password = models.CharField(max_length=255)
  birthdate = models.DateField()
  registered_at = models.DateTimeField()



import json
from inertia.test import InertiaTestCase, inertia_page, inertia_div
from django.test import override_settings
from unittest.mock import patch, Mock
from requests.exceptions import RequestException

@override_settings(
  INERTIA_SSR_ENABLED=True,
  INERTIA_SSR_URL='ssr-url',
  INERTIA_VERSION='1.0',
)
class SSRTestCase(InertiaTestCase):

  @patch('inertia.http.requests')
  def test_it_returns_ssr_calls(self, mock_request):
    mock_response = Mock()
    mock_response.json.return_value = {
      'body': '<div>Body Works</div>',
      'head': '<title>Head works</title>',
    }

    mock_request.post.return_value = mock_response
    
    response = self.client.get('/props/')
    
    mock_request.post.assert_called_once_with(
      'ssr-url/render',
      data=json.dumps(inertia_page('props', props={'name': 'Brandon', 'sport': 'Hockey'})),
      headers={'Content-Type': 'application/json'},
    )
    self.assertTemplateUsed('inertia_ssr.html')
    print(response.content)
    self.assertContains(response, '<div>Body Works</div>')
    self.assertContains(response, 'head--<title>Head works</title>--head')


  @patch('inertia.http.requests')
  def test_it_uses_inertia_if_inertia_requests_are_made(self, mock_requests):
    response = self.inertia.get('/props/')

    mock_requests.post.assert_not_called()
    self.assertJSONResponse(response, inertia_page('props', props={'name': 'Brandon', 'sport': 'Hockey'}))

  @patch('inertia.http.requests')
  def test_it_fallsback_on_failure(self, mock_requests):
    def uh_oh(*args, **kwargs):
      raise RequestException()

    mock_response = Mock()
    mock_response.raise_for_status.side_effect = uh_oh
    mock_requests.post.return_value = mock_response

    response = self.client.get('/props/')
    self.assertContains(response, inertia_div('props', props={'name': 'Brandon', 'sport': 'Hockey'}))



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

INERTIA_LAYOUT = 'layout.html'

MIDDLEWARE = [
  'django.middleware.security.SecurityMiddleware',
  'django.contrib.sessions.middleware.SessionMiddleware',
  'django.middleware.common.CommonMiddleware',
  'django.middleware.csrf.CsrfViewMiddleware',
  'django.contrib.auth.middleware.AuthenticationMiddleware',
  'django.contrib.messages.middleware.MessageMiddleware',
  'django.middleware.clickjacking.XFrameOptionsMiddleware',
  'inertia.middleware.InertiaMiddleware',
]

INSTALLED_APPS = [
  "django.contrib.admin",
  "django.contrib.auth",
  "django.contrib.contenttypes",
  "django.contrib.sessions",
  "django.contrib.sites",
  "inertia",
  "inertia.tests.testapp.apps.TestAppConfig"
]

ROOT_URLCONF = 'inertia.tests.testapp.urls'

TEMPLATES = [
  {
    'BACKEND': 'django.template.backends.django.DjangoTemplates',
    'DIRS': [
      BASE_DIR / 'tests/testapp',
    ],
    'APP_DIRS': True,
    'OPTIONS': {
      'context_processors': [
        'django.template.context_processors.debug',
        'django.template.context_processors.request',
        'django.contrib.auth.context_processors.auth',
        'django.contrib.messages.context_processors.messages',
      ],
    },
  },
]

# only in place to silence an error
SECRET_KEY = 'django-insecure-3p_!uve+em7f45+74jh16)y)h00ve@9d2edh=cuebdsrbco%vb'
DATABASES = {
  "default": {
    "ENGINE": "django.db.backends.sqlite3",
    "NAME": "unused"
  }
}

# silence a warning
USE_TZ = False



from inertia.test import InertiaTestCase, inertia_div, inertia_page
from pytest import warns

class FirstLoadTestCase(InertiaTestCase):
  def test_with_props(self):
    self.assertContains(
      self.client.get('/props/'),
      inertia_div('props', props={
        'name': 'Brandon',
        'sport': 'Hockey',
      })
    )

  def test_with_template_data(self):
    self.assertContains(
      self.client.get('/template_data/'),
      inertia_div('template_data', template_data={
        'name': 'Brian',
        'sport': 'Basketball',
      })
    )

  def test_with_no_data(self):
    self.assertContains(
      self.client.get('/empty/'),
      inertia_div('empty')
    )

  def test_proper_status_code(self):
    self.assertEqual(
      self.client.get('/empty/').status_code,
      200
    )

  def test_template_rendered(self):
    self.assertTemplateUsed(self.client.get('/empty/'), 'inertia.html')


class SubsequentLoadTestCase(InertiaTestCase):
  def test_with_props(self):
    self.assertJSONResponse(
      self.inertia.get('/props/'),
      inertia_page('props', props={
        'name': 'Brandon',
        'sport': 'Hockey',
      })
    )

  def test_with_template_data(self):
    self.assertJSONResponse(
      self.inertia.get('/template_data/'),
      inertia_page('template_data', template_data={
        'name': 'Brian',
        'sport': 'Basketball',
      })
    )

  def test_with_no_data(self):
    self.assertJSONResponse(
      self.inertia.get('/empty/'),
      inertia_page('empty')
    )

  def test_proper_status_code(self):
    self.assertEqual(
      self.inertia.get('/empty/').status_code,
      200
    )

  def test_redirects_from_inertia_views(self):
    self.assertEqual(
      self.inertia.get('/inertia-redirect/').status_code,
      302
    )

class LazyPropsTestCase(InertiaTestCase):
  def test_lazy_props_are_not_included(self):
    with warns(DeprecationWarning):
      self.assertJSONResponse(
        self.inertia.get('/lazy/'),
        inertia_page('lazy', props={'name': 'Brian'})
      )

  def test_lazy_props_are_included_when_requested(self):
    with warns(DeprecationWarning):
      self.assertJSONResponse(
        self.inertia.get('/lazy/', HTTP_X_INERTIA_PARTIAL_DATA='sport,grit', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
        inertia_page('lazy', props={'sport': 'Basketball', 'grit': 'intense'})
      )

class OptionalPropsTestCase(InertiaTestCase):
  def test_optional_props_are_not_included(self):
    self.assertJSONResponse(
      self.inertia.get('/optional/'),
      inertia_page('optional', props={'name': 'Brian'})
    )

  def test_optional_props_are_included_when_requested(self):
    self.assertJSONResponse(
      self.inertia.get('/optional/', HTTP_X_INERTIA_PARTIAL_DATA='sport,grit', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
      inertia_page('optional', props={'sport': 'Basketball', 'grit': 'intense'})
    )

class ComplexPropsTestCase(InertiaTestCase):
  def test_nested_callable_props_work(self):
    self.assertJSONResponse(
      self.inertia.get('/complex-props/'),
      inertia_page('complex-props', props={'person': {'name': 'Brandon'}})
    )

class ShareTestCase(InertiaTestCase):
  def test_that_shared_props_are_merged(self):
    self.assertJSONResponse(
      self.inertia.get('/share/'),
      inertia_page('share', props={'name': 'Brandon', 'position': 'goalie', 'number': 29})
    )

class CSRFTestCase(InertiaTestCase):
  def test_that_csrf_inclusion_is_automatic(self):
    response = self.inertia.get('/props/')

    self.assertIsNotNone(response.cookies.get('csrftoken'))

  def test_that_csrf_is_included_even_on_initial_page_load(self):
    response = self.client.get('/props/')

    self.assertIsNotNone(response.cookies.get('csrftoken'))

class DeferredPropsTestCase(InertiaTestCase):
  def test_deferred_props_are_set(self):
    self.assertJSONResponse(
      self.inertia.get('/defer/'),
      inertia_page(
        'defer', 
        props={'name': 'Brian'}, 
        deferred_props={'default': ['sport']})
    )

  def test_deferred_props_are_grouped(self):
    self.assertJSONResponse(
      self.inertia.get('/defer-group/'),
      inertia_page(
        'defer-group', 
        props={'name': 'Brian'}, 
        deferred_props={'group': ['sport', 'team'], 'default': ['grit']}) 
    )

  def test_deferred_props_are_included_when_requested(self):
    self.assertJSONResponse(
      self.inertia.get('/defer/', HTTP_X_INERTIA_PARTIAL_DATA='sport', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
      inertia_page('defer', props={'sport': 'Basketball'})
    )
  

  def test_only_deferred_props_in_group_are_included_when_requested(self):
    self.assertJSONResponse(
      self.inertia.get('/defer-group/', HTTP_X_INERTIA_PARTIAL_DATA='sport,team', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
      inertia_page('defer-group', props={'sport': 'Basketball', 'team': 'Bulls'})
    )

    self.assertJSONResponse(
      self.inertia.get('/defer-group/', HTTP_X_INERTIA_PARTIAL_DATA='grit', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
      inertia_page('defer-group', props={'grit': 'intense'})
    )

class MergePropsTestCase(InertiaTestCase):
  def test_merge_props_are_included_on_initial_load(self):
    self.assertJSONResponse(
      self.inertia.get('/merge/'),
      inertia_page('merge', props={
        'name': 'Brandon',
        'sport': 'Hockey',
      }, merge_props=['sport', 'team'], deferred_props={'default': ['team']})
    )


  def test_deferred_merge_props_are_included_on_subsequent_load(self):
    self.assertJSONResponse(
      self.inertia.get('/merge/', HTTP_X_INERTIA_PARTIAL_DATA='team', HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent'),
      inertia_page('merge', props={
        'team': 'Penguins',
      }, merge_props=['sport', 'team'])
    )
  
  def test_merge_props_are_not_included_when_reset(self):
    self.assertJSONResponse(
      self.inertia.get(
        '/merge/',
        HTTP_X_INERTIA_PARTIAL_DATA='sport,team',
        HTTP_X_INERTIA_PARTIAL_COMPONENT='TestComponent',
        HTTP_X_INERTIA_RESET='sport,team'
      ),
      inertia_page('merge', props={
        'sport': 'Hockey',
        'team': 'Penguins',
      })
    )


from django.test import TestCase, Client, override_settings
from inertia.test import InertiaTestCase

class MiddlewareTestCase(InertiaTestCase):
  def test_anything(self):
    response = self.client.get('/test/')

    self.assertEqual(response.status_code, 200)

  def test_stale_versions_are_refreshed(self):
    response = self.inertia.get('/empty/',
      HTTP_X_INERTIA_VERSION='some-nonsense',
    )

    self.assertEqual(response.status_code, 409)
    self.assertEqual(response.headers['X-Inertia-Location'], 'http://testserver/empty/')

  def test_redirect_status(self):
    response = self.inertia.post('/redirect/')
    self.assertEqual(response.status_code, 302)

    for http_method in ['put', 'patch', 'delete']:
      response = getattr(self.inertia, http_method)('/redirect/')

      self.assertEqual(response.status_code, 303)

  def test_a_request_not_from_inertia_is_ignored(self):
    response = self.client.get('/empty/',
      HTTP_X_INERTIA_VERSION='some-nonsense',
    )

    self.assertEqual(response.status_code, 200)

  def test_external_redirect_status(self):
    response = self.inertia.post('/external-redirect/')
    self.assertEqual(response.status_code, 409)
    self.assertIn("X-Inertia-Location", response.headers)
    self.assertEqual(
      "http://foobar.com/", response.headers["X-Inertia-Location"]
    )



from inertia.http import encrypt_history
from inertia.test import InertiaTestCase
from django.test import override_settings
from django.test.client import RequestFactory
from inertia.tests.testapp.views import empty_test


class MiddlewareTestCase(InertiaTestCase):
  def test_encrypt_history_setting(self):
    self.client.get('/empty/')
    assert self.page()['encryptHistory'] is False

    with override_settings(INERTIA_ENCRYPT_HISTORY=True):
      self.client.get('/empty/')
      assert self.page()['encryptHistory'] is True

  def test_encrypt_history(self):
    self.client.get('/encrypt-history/')
    assert self.page()['encryptHistory'] is True

    with override_settings(INERTIA_ENCRYPT_HISTORY=True):
      self.client.get('/no-encrypt-history/')
      assert self.page()['encryptHistory'] is False

  def test_clear_history(self):
    self.client.get('/clear-history/')
    assert self.page()['clearHistory'] is True

  def test_clear_history_redirect(self):
    response = self.client.get('/clear-history-redirect/', follow=True)
    self.assertRedirects(response, '/empty/')
    assert self.page()['clearHistory'] is True

  def test_raises_type_error(self):
    with self.assertRaisesMessage(TypeError, 'Expected boolean for encrypt_history, got str'):
      self.client.get('/encrypt-history-type-error/')

    with self.assertRaisesMessage(TypeError, 'Expected boolean for clear_history, got str'):
      self.client.get('/clear-history-type-error/')



from django.test import TestCase, Client
from unittest.mock import patch
from django.http.response import JsonResponse
from inertia.settings import settings
from json import dumps, loads
from django.utils.html import escape
from django.shortcuts import render

class BaseInertiaTestCase:
  def setUp(self):
    self.inertia = Client(HTTP_X_INERTIA=True)

  def assertJSONResponse(self, response, json_obj):
    self.assertIsInstance(response, JsonResponse)
    self.assertEqual(response.json(), json_obj)

class InertiaTestCase(BaseInertiaTestCase, TestCase):
  def setUp(self):
    super().setUp()

    self.mock_inertia = patch('inertia.http.base_render', wraps=render)
    self.mock_render = self.mock_inertia.start()

  def tearDown(self):
    self.mock_inertia.stop()

  def page(self):
    return loads(self.mock_render.call_args.args[2]['page'])

  def props(self):
    return self.page()['props']
  
  def merge_props(self):
    return self.page()['mergeProps']
  
  def deferred_props(self):
    return self.page()['deferredProps']

  def template_data(self):
    context = self.mock_render.call_args.args[2]

    return {key: context[key] for key in context if key not in ['page', 'inertia_layout']}

  def component(self):
    return self.page()['component']

  def assertIncludesProps(self, props):
    self.assertDictEqual(self.props(), {**self.props(), **props})

  def assertHasExactProps(self, props):
    self.assertDictEqual(self.props(), props)

  def assertIncludesTemplateData(self, template_data):
    self.assertDictEqual(self.template_data(), {**self.template_data(), **template_data})

  def assertHasExactTemplateData(self, template_data):
    self.assertDictEqual(self.template_data(), template_data)

  def assertComponentUsed(self, component_name):
    self.assertEqual(component_name, self.component())

def inertia_page(url, component='TestComponent', props={}, template_data={}, deferred_props=None, merge_props=None):
  _page = {
    'component': component,
    'props': props,
    'url': f'http://testserver/{url}/',
    'version': settings.INERTIA_VERSION,
    'encryptHistory': False,
    'clearHistory': False,
  }

  if deferred_props:
    _page['deferredProps'] = deferred_props
  
  if merge_props:
    _page['mergeProps'] = merge_props
    
  return _page

def inertia_div(*args, **kwargs):
  page = inertia_page(*args, **kwargs)
  return f'<div id="app" data-page="{escape(dumps(page))}"></div>'



from inertia.test import InertiaTestCase
from django.test import override_settings

class SettingsTestCase(InertiaTestCase):
  @override_settings(
    INERTIA_VERSION='2.0'
  )
  def test_version_works(self):
    response = self.inertia.get('/empty/', HTTP_X_INERTIA_VERSION='2.0')

    self.assertEqual(response.status_code, 200)

  def test_version_fallsback(self):
    response = self.inertia.get('/empty/', HTTP_X_INERTIA_VERSION='1.0')

    self.assertEqual(response.status_code, 200)

  def test_layout(self):
    response = self.client.get('/empty/')
    self.assertTemplateUsed(response, 'layout.html')



from django.conf import settings as django_settings
from .utils import InertiaJsonEncoder

__all__ = ['settings']

class InertiaSettings:
  INERTIA_VERSION = '1.0'
  INERTIA_JSON_ENCODER = InertiaJsonEncoder
  INERTIA_SSR_URL = 'http://localhost:13714'
  INERTIA_SSR_ENABLED = False
  INERTIA_ENCRYPT_HISTORY = False
  
  def __getattribute__(self, name):
    try:
      return getattr(django_settings, name)
    except AttributeError:
      return super().__getattribute__(name)

settings = InertiaSettings()



from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models.query import QuerySet
from django.forms.models import model_to_dict as base_model_to_dict
from .prop_classes import OptionalProp, DeferredProp, MergeProp
import warnings

def model_to_dict(model):
  return base_model_to_dict(model, exclude=('password',))

class InertiaJsonEncoder(DjangoJSONEncoder):
  def default(self, value):
    if isinstance(value, models.Model):
      return model_to_dict(value)
    
    if isinstance(value, QuerySet):
      return [model_to_dict(model) for model in value]
    
    return super().default(value)

def lazy(prop):
  warnings.warn(
    "lazy is deprecated and will be removed in a future version. Please use optional instead.",
    DeprecationWarning,
    stacklevel=2
  )
  return optional(prop)

def optional(prop):
  return OptionalProp(prop)

def defer(prop, group="default", merge=False):
  return DeferredProp(prop, group=group, merge=merge)

def merge(prop):
  return MergeProp(prop)



__all__ = ['share']

class InertiaShare:
  def __init__(self):
    self.props = {}

  def set(self, **kwargs):
    self.props = {
      **self.props,
      **kwargs,
    }

  def all(self):
    return self.props


def share(request, **kwargs):
  if not hasattr(request, 'inertia'):
    request.inertia = InertiaShare()

  request.inertia.set(**kwargs)



from .settings import settings
from django.contrib import messages
from django.http import HttpResponse
from django.middleware.csrf import get_token
from .http import location

class InertiaMiddleware:
  def __init__(self, get_response):
    self.get_response = get_response
  
  def __call__(self, request):
    response = self.get_response(request)

    # Inertia requests don't ever render templates, so they skip the typical Django
    # CSRF path. We'll manually add a CSRF token for every request here.
    get_token(request)

    if not self.is_inertia_request(request):
      return response

    if self.is_non_post_redirect(request, response):
      response.status_code = 303

    if self.is_stale(request):
      return self.force_refresh(request)

    return response

  def is_non_post_redirect(self, request, response):
    return self.is_redirect_request(response) and request.method in ['PUT', 'PATCH', 'DELETE']

  def is_inertia_request(self, request):
    return 'X-Inertia' in request.headers

  def is_redirect_request(self, response):
    return response.status_code in [301, 302]

  def is_stale(self, request):
    return request.headers.get('X-Inertia-Version', settings.INERTIA_VERSION) != settings.INERTIA_VERSION

  def is_stale_inertia_get(self, request):
    return request.method == 'GET' and self.is_stale(request)

  def force_refresh(self, request):
    messages.get_messages(request).used = False
    return location(request.build_absolute_uri())



from abc import ABC, abstractmethod
import warnings

class CallableProp(ABC):
  def __init__(self, prop):
    self.prop = prop
  
  def __call__(self):
    return self.prop() if callable(self.prop) else self.prop

class MergeableProp(ABC):
  @abstractmethod
  def should_merge(self):
    pass

class IgnoreOnFirstLoadProp(ABC):
  pass

class OptionalProp(CallableProp, IgnoreOnFirstLoadProp):
  pass

class DeferredProp(CallableProp, MergeableProp, IgnoreOnFirstLoadProp):
  def __init__(self, prop, group, merge=False):
    super().__init__(prop)
    self.group = group
    self.merge = merge
  
  def should_merge(self):
    return self.merge

class MergeProp(CallableProp, MergeableProp):
  def should_merge(self):
    return True
